/*
	$Header$
	$Log$ 
 */

package COM.webconcepts.net;

import java.net.*;
import java.io.*;
import java.util.*;

/** 
  * MultiThreadServer takes some of the drudgery out of writing network servers.
  * It uses Java's metaclassing abilities to automatically create instances of service
  * thread classes for a thread pool without knowing their type until run-time. Thus any
  * threads that extend ServiceThread can be used to handle incoming clients; in theory
  * it is possible to write a server without adding any code to MultiThreadServer except
  * a main() method to run the server itself. You can concentrate on writing the protocol-
  * specific client-handlign code and ignore the rest for the most part
  *
  * MultiThreadServer relies on a fixed, non-regenerating pool of threads. One possible
  * line of improvement would be to add a monitor thread that watches for dying threads
  * and replaces them with new ones.
  * 
  * @author Kyle Downey
  * @author Copyright 1997, WebConcepts, LLC, http://www.webconcepts.com
  * @author This code is free for use under the terms of the GNU General Public License
  * @author http://www.fsf.org
  *
  * @see COM.webconcepts.net.HttpProxyServer
  * @see COM.webconcepts.net.ServiceThread
  * @see COM.webconcepts.net.MultiThreadServer
  */
public class MultiThreadServer {
	private static int DEFAULT_THREADPOOL_SIZE = 5;
	private static int DEFAULT_BACKLOG_SIZE = 5;
	private ServerSocket ss;
	private int servicePort;
	private int threadPoolSize;
	private int backlogSize;
	private Vector threadPool;
	private Class serviceThreadClass;

	/** 
	  * creates a multi-threaded server with a default backlog size and thread pool size (5)
	  * @param serviceThreadClass the class to instantiate when creating threads
	  * @param servicePort the port to listen on for clients
	  */
	public MultiThreadServer(Class serviceThreadClass, int servicePort) {
		this.serviceThreadClass = serviceThreadClass;
		this.servicePort = servicePort;
		this.threadPoolSize = DEFAULT_THREADPOOL_SIZE;
		this.backlogSize = DEFAULT_BACKLOG_SIZE;
	}

	/** 
	  * creates a multi-threaded server with a default backlog size (5)
	  * @param serviceThreadClass the class to instantiate when creating threads
	  * @param servicePort the port to listen on for clients
	  * @param threadPoolSize the number of service threads to create
	  */
	public MultiThreadServer(Class serviceThreadClass, int servicePort, int threadPoolSize) {
		this.serviceThreadClass = serviceThreadClass;
		this.servicePort = servicePort;
		this.threadPoolSize = threadPoolSize;
		this.backlogSize = DEFAULT_BACKLOG_SIZE;
	}

	/** 
	  * creates a multi-threaded server
	  * @param serviceThreadClass the class to instantiate when creating threads
	  * @param servicePort the port to listen on for clients
	  * @param threadPoolSize the number of service threads to create
	  * @param backlogSize the maximum number of clients to keep waiting for a free thread
	  */
	public MultiThreadServer(Class serviceThreadClass, int servicePort, int threadPoolSize, int backlogSize) {
		this.serviceThreadClass = serviceThreadClass;
		this.servicePort = servicePort;
		this.threadPoolSize = threadPoolSize;
		this.backlogSize = backlogSize;
	}

	/**
	  * init() populates the thread pool and starts each listener thread
	  * @exception IOException if there are network errors with the server socket
	  * @exception InstantiationException if there are problems creating service threads
	  * @exception IllegalAccessException if there are problems finding the service thread
	  * class
	  */
	public void init() throws IOException, InstantiationException, IllegalAccessException {
		threadPool = new Vector();
		ss = new ServerSocket(servicePort, backlogSize);

		if (!(serviceThreadClass.newInstance() instanceof ServiceThread)) {
			throw new InstantiationException("service thread class must extend ServiceThread");
		}

		int i;
		for (i = 0; i < threadPoolSize; i++) {
			ServiceThread serviceThread = (ServiceThread)serviceThreadClass.newInstance();
			threadPool.addElement(serviceThreadClass.newInstance());
			serviceThread.setServer(this);
			serviceThread.start();
		}
	}

	/**
	  * returns the server socket
	  * @return a reference to the server socket
	  */
	public ServerSocket getServerSocket() {
		return this.ss;
	}
}